==> acst-0.1.1.tar.gz: acst-0.1.1/acst.h <==
/**
 * Copyright (C) 2022 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef ACST_H
#define ACST_H

#include <stdbool.h>        // for bool
#include <stdio.h>          // for printf
#include <stdlib.h>         // for exit, EXIT_SUCCESS, size_t

#define BUFSZ               8192    /* buffer size to use in hash computation */
#define SHA256_BYTES        32      /* byte size of SHA256 hash */
#define CSSZ                64      /* size of SHA256 hash stored as hex */
#define TSSZ                32      /* allocation size for timespan holder */
#define XATTR_CS            "user.acst.cs"  /* name of checksum xattr */
#define XATTR_TS            "user.acst.ts"  /* name of timestamp xattr */
#define FRMT_TS             "%010llu.%09lu" /* format of timestamp xattr */

/* string values used in program, could be changes for language compliance */
#define STR_CMP_FRMT        " stored: %s "FRMT_TS"\n"                       \
                            " actual: %s "FRMT_TS"\n"
#define STR_ERR_OOM         "Error: out of memory\n"
#define STR_ERR_ABNO        "Error: abnormal changes detected \"%s\"\n"
#define STR_ERR_OPNF        "Error: could not open file \"%s\"\n"
#define STR_ERR_OPND        "Error: could not open directory \"%s\"\n"
#define STR_ERR_DDEV        "Error: \"%s\" belong to a different devices than \"%s\"\n"
#define STR_ERR_REGF        "Error: not a regular file \"%s\"\n"
#define STR_ERR_HASH        "Error: could not compute hash \"%s\"\n"
#define STR_ERR_XARM        "Error: could not remove extended attributes from file \"%s\": %s\n"
#define STR_ERR_XAWR        "Error: could not write extended attributes to file \"\%s\"\n"
#define STR_ERR_RECU        "Error: path is a directory, did you mean to use the -r option? \"%s\"\n"
#define STR_OUT_STAT        "<%s> %s\n"
#define STR_OUT_NEW         "new"
#define STR_OUT_OK          "ok"
#define STR_OUT_TOUC        "hash ok"
#define STR_OUT_OUTD        "outdated"
#define STR_OUT_BACK        "backdated"
#define STR_OUT_CORR        "corrupt"
#define STR_OUT_MALF        "malformed"
#define STR_OUT_DISR        "disrupted"
#define STR_OUT_REMO        "xattr removed"

#define USAGE(code)                                                         \
        {                                                                   \
          printf("Usage: %s [OPTION]... <FILE>...\n", arg.prg);             \
          if ((code) == EXIT_SUCCESS)                                       \
            printf(                                                         \
    "\n"                                                                    \
    "Tracks changes and corruption in files using xattr-based checksums.\n" \
    "\n"                                                                    \
    "Positional arguments:\n"                                               \
    "  FILE                  file(s) to track\n"                            \
    "\n"                                                                    \
    "Optional arguments:\n"                                                 \
    "  -h                    show this help message and exit\n"             \
    "  -m                    summarize information at end of execution\n"   \
    "  -n                    don't create or update any file attributes\n"  \
    "  -q                    quiet mode, lowers verbosity\n"                \
    "  -r                    process files and directories recursively\n"   \
    "  -V                    output version information and exit\n"         \
    "  -x                    remove extended attributes from file(s)\n"     \
    "\n"                                                                    \
    "Full documentation <https://noxz.tech/software/acst>\n"                \
            );                                                              \
          exit(code);                                                       \
        }
#define VER()                                                               \
        {                                                                   \
          printf("%s " VERSION "\n", arg.prg);                              \
          exit(EXIT_SUCCESS);                                               \
        }
#define SUMMERIZE()                                                         \
        {                                                                   \
          if (arg.remove)                                                   \
            printf(         "\n"                                            \
                            "Total files processed      : %d\n"             \
                            "  Xattr removed            : %d\n"             \
                            "Total errors               : %d\n"             \
                            "  Opening files            : %d\n"             \
                            "  Non-regular files        : %d\n"             \
                            "  Xattr operations         : %d\n"             \
                            "  Generic                  : %d\n"             \
              , cnt.total, cnt.ok                                           \
              , cnt.errs, cnt.errOpening, cnt.errNotRegular                 \
              , cnt.errWritingXattr, cnt.errGeneric);                       \
          else                                                              \
            printf(         "\n"                                            \
                            "Total files processed      : %d\n"             \
                            "  New                      : %d\n"             \
                            "  Ok                       : %d\n"             \
                            "  Hash ok                  : %d\n"             \
                            "  Outdated                 : %d\n"             \
                            "  Backdated                : %d\n"             \
                            "  Corrupt                  : %d\n"             \
                            "  Malformed                : %d\n"             \
                            "  Disrupted                : %d\n"             \
                            "Total errors               : %d\n"             \
                            "  Opening files            : %d\n"             \
                            "  Non-regular files        : %d\n"             \
                            "  Xattr operations         : %d\n"             \
                            "  Generic                  : %d\n"             \
              , cnt.total, cnt.new, cnt.ok, cnt.touched                     \
              , cnt.outdated, cnt.backdated, cnt.corrupt                    \
              , cnt.malformed, cnt.disrupted                                \
              , cnt.errs, cnt.errOpening, cnt.errNotRegular                 \
              , cnt.errWritingXattr, cnt.errGeneric);                       \
        }

/* helper variables used by various functions */
static char ts[TSSZ];                               /* timestamp holder */
static const char hextab[]  = "0123456789abcdef";   /* hex digit table */
static const char zSHA256[] = "0000000000000000"    /* zeroed out SHA256 */
                              "0000000000000000"    /* why 4 lines?... */
                              "0000000000000000"
                              "0000000000000000";   /* ...it looks better! */

enum Error {
	ER_NOT_REGULAR,         /* not a regular file error */
	ER_OPENING,             /* error opening file or directory */
	ER_XATTR_OPERATION,     /* error when performing xattr operations */
	ER_GENERIC,             /* generic error */
	ER_FATAL                /* fatal error, should result in program ending */
};

enum FileState {
	FS_OK,                  /* checksum and mtime both match */
	FS_DISRUPTED,           /* file was changed during hash computation */
	FS_CORRUPT,             /* checksum differs while mtime matches */
	FS_MALFORMED,           /* xattrs cannot be read in its intended format */
	FS_TOUCHED,             /* checksum matches but mtime doesn't */
	FS_NEW,                 /* file has no prior checksum or mtime stored */
	FS_OUTDATED,            /* checksum differs and mtime is newer */
	FS_BACKDATED,           /* checksum differs and mtime is older */
	FS_REMOVED,             /* xattrs have been removed from file */
	FS_REMOVED_ERROR,       /* removal of xattrs failed */
	FS_ERROR                /* general error */
};

/* extended attribute/metadata structure */
typedef struct {
	unsigned long long s;   /* seconds */
	unsigned long ns;       /* nanoseconds */
	char cs[CSSZ + 1];      /* checksum */
	int tcmp;               /* timespan comparison holder */
} xa_t;

/* arguments collected from command line */
struct Arguments {
	const char *prg;        /* program name */
	bool dryrun;            /* make a dry run */
	int quiet;              /* level of quietness */
	bool recursive;         /* indicate to open and walk directories */
	bool remove;            /* remove xattrs */
	bool summarize;         /* show summery at end of program */
} arg;

/* counters */
struct Counters {
	int errs;               /* all errors */
	int errNotRegular;      /* not a regular file errors  */
	int errOpening;         /* errors opening file or directory */
	int errWritingXattr;    /* errors when performing xattr operations */
	int errGeneric;         /* generic errors */

	int total;              /* all non error counters */
	int ok;                 /* checksum and mtime both match */
	int disrupted;          /* file was changed during hash computation */
	int corrupt;            /* checksum differs while mtime matches */
	int malformed;          /* xattrs cannot be read in its intended format */
	int outdated;           /* checksum differs and mtime is newer */
	int backdated;          /* checksum differs and mtime is older */
	int touched;            /* checksum matches but mtime doesn't */
	int new;                /* file has no prior checksum or mtime stored */
} cnt;

/**
 * Converts binary data into hex format.
 *
 * @param bin               The binary data
 * @param len               The size of the data
 * @param hex               The result of the converted data
 */
static void bin2hex(unsigned char *bin, size_t len, char *hex);

/**
 * Retrieves and reports error types and messages. If error type is of fatal
 * nature, the program will exit with EXIT_FAILURE.
 *
 * @param er                The type of error (see enum Error above)
 * @param fmt               The output format of the error message
 * @param ...               Various arguments to output in the message
 */
static void error(enum Error er, const char *fmt, ...);

/**
 * Final stage of processing a file which returns the state of the file being
 * processed meanwhile metadata holders are being populated.
 *
 * @param fd                File descriptor of the file
 * @param xa_s              Metadata holder of (stored) extended attributes
 * @param xa_a              Metadata holder of (actual) extended attributes
 *
 * @returns                 The file state of the file being processed
 */
static enum FileState file_state(int fd, xa_t *xa_s, xa_t *xa_a);

/**
 * Processes arguments (file names) either from command line or from directory
 * output.
 *
 * @param fn                File name of the file
 */
static void process_argument(const char *fn);

/**
 * Processes directories if stated to do so (see arg.recursive) by looping
 * through their content (children).
 *
 * @param fn                File name of the directory
 */
static void process_directory(const char *fn);

/**
 * Processes files and returns either error messages or other output if stated
 * to do so (see arg.quiet) based on their file state (see enum FileState
 * above).
 *
 * @param fn                File name of the file
 */
static void process_file(const char *fn);

/**
 * Reports on files and their file state.
 *
 * @param fn                File name of the file
 * @param st                The state of the file
 */
static void report(const char *fn, enum FileState st);

/**
 * Compute and retrieve metadata for the checksum based on file content
 * together with the date modified.
 *
 * @param fd                File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred
 */
static bool xa_compute(int fd, xa_t *xa);

/**
 * Computes the hash checksum of a file based on the SHA256 hashing algorithm.
 *
 * @param fd                File descriptor of the file
 * @param cs                Pointer for the checksum
 */
static void xa_hash(int fd, char *cs);

/**
 * Reads and retrieves extended attributes as metadata from a file.
 *
 * @param fd                File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred (malformed data)
 */
static bool xa_read(int fd, xa_t *xa);

/**
 *  Writes, or removes if stated (see arg.remove), extended attributes as
 *  metadata to a file.
 *
 * @param fd               File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred (write error)
 */
static bool xa_write(int fd, const xa_t *xa);

#endif /* !ACST_H */

==> acst-0.1.3.tar.gz: acst-0.1.3/acst.h <==
/**
 * Copyright (C) 2022 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef ACST_H
#define ACST_H

#include <stdbool.h>        // for bool
#include <stdio.h>          // for printf
#include <stdlib.h>         // for exit, EXIT_SUCCESS, size_t

#define BUFSZ               8192    /* buffer size to use in hash computation */
#define SHA256_BYTES        32      /* byte size of SHA256 hash */
#define CSSZ                64      /* size of SHA256 hash stored as hex */
#define TSSZ                32      /* allocation size for timespan holder */
#define XATTR_CS            "user.acst.cs"  /* name of checksum xattr */
#define XATTR_TS            "user.acst.ts"  /* name of timestamp xattr */
#define FRMT_TS             "%010llu.%09lu" /* format of timestamp xattr */

/* string values used in program, could be changes for language compliance */
#define STR_CMP_FRMT        " stored: %s "FRMT_TS"\n"                       \
                            " actual: %s "FRMT_TS"\n"
#define STR_DPN_FRMT        " %s\n"
#define STR_DPH_FRMT        "<dup> %s\n"STR_DPN_FRMT
#define STR_DPS_FRMT        "Found %d files with duplicates.\n"
#define STR_ERR_OOM         "Error: out of memory\n"
#define STR_ERR_PMAX        "Error: PATH_MAX reached before end\n"
#define STR_ERR_ABNO        "Error: abnormal changes detected \"%s\"\n"
#define STR_ERR_OPNF        "Error: could not open file \"%s\"\n"
#define STR_ERR_OPND        "Error: could not open directory \"%s\"\n"
#define STR_ERR_DDEV        "Error: \"%s\" belong to a different devices than \"%s\"\n"
#define STR_ERR_REGF        "Error: not a regular file \"%s\"\n"
#define STR_ERR_HASH        "Error: could not compute hash \"%s\"\n"
#define STR_ERR_XARM        "Error: could not remove extended attributes from file \"%s\": %s\n"
#define STR_ERR_XAWR        "Error: could not write extended attributes to file \"\%s\"\n"
#define STR_ERR_RECU        "Error: path is a directory, did you mean to use the -r option? \"%s\"\n"
#define STR_OUT_STAT        "<%s> %s\n"
#define STR_OUT_NEW         "new"
#define STR_OUT_OK          "ok"
#define STR_OUT_TOUC        "hash ok"
#define STR_OUT_OUTD        "outdated"
#define STR_OUT_BACK        "backdated"
#define STR_OUT_CORR        "corrupt"
#define STR_OUT_MALF        "malformed"
#define STR_OUT_DISR        "disrupted"
#define STR_OUT_REMO        "xattr removed"

#define MIN(X, Y)           (((X) < (Y)) ? (X) : (Y))
#define USAGE(code)                                                         \
        {                                                                   \
          printf("Usage: %s [OPTION]... <FILE>...\n", gl.prg);              \
          if ((code) == EXIT_SUCCESS)                                       \
            printf(                                                         \
    "\n"                                                                    \
    "Tracks changes and corruption in files using xattr-based checksums.\n" \
    "\n"                                                                    \
    "Positional arguments:\n"                                               \
    "  FILE                  file(s) to track\n"                            \
    "\n"                                                                    \
    "Optional arguments:\n"                                                 \
    "  -h                    show this help message and exit\n"             \
    "  -m                    summarize information at end of execution\n"   \
    "  -n                    don't create or update any file attributes\n"  \
    "  -q                    quiet mode, lowers verbosity\n"                \
    "  -r                    process files and directories recursively\n"   \
    "  -V                    output version information and exit\n"         \
    "  -x                    remove extended attributes from file(s)\n"     \
    "\n"                                                                    \
    "Full documentation <https://noxz.tech/software/acst>\n"                \
            );                                                              \
          exit(code);                                                       \
        }
#define VER()                                                               \
        {                                                                   \
          printf("%s " VERSION "\n", gl.prg);                               \
          exit(EXIT_SUCCESS);                                               \
        }
#define SUMMERIZE()                                                         \
        {                                                                   \
          if (arg.remove)                                                   \
            printf(         "\n"                                            \
                            "Total files processed      : %d\n"             \
                            "  Xattr removed            : %d\n"             \
                            "Total errors               : %d\n"             \
                            "  Opening files            : %d\n"             \
                            "  Non-regular files        : %d\n"             \
                            "  Xattr operations         : %d\n"             \
                            "  Generic                  : %d\n"             \
              , cnt.total, cnt.ok                                           \
              , cnt.errs, cnt.errOpening, cnt.errNotRegular                 \
              , cnt.errWritingXattr, cnt.errGeneric);                       \
          else                                                              \
            printf(         "\n"                                            \
                            "Total files processed      : %d\n"             \
                            "  New                      : %d\n"             \
                            "  Ok                       : %d\n"             \
                            "  Hash ok                  : %d\n"             \
                            "  Outdated                 : %d\n"             \
                            "  Backdated                : %d\n"             \
                            "  Corrupt                  : %d\n"             \
                            "  Malformed                : %d\n"             \
                            "  Disrupted                : %d\n"             \
                            "Total errors               : %d\n"             \
                            "  Opening files            : %d\n"             \
                            "  Non-regular files        : %d\n"             \
                            "  Xattr operations         : %d\n"             \
                            "  Generic                  : %d\n"             \
              , cnt.total, cnt.new, cnt.ok, cnt.touched                     \
              , cnt.outdated, cnt.backdated, cnt.corrupt                    \
              , cnt.malformed, cnt.disrupted                                \
              , cnt.errs, cnt.errOpening, cnt.errNotRegular                 \
              , cnt.errWritingXattr, cnt.errGeneric);                       \
        }

/* helper variables used by various functions */
static char ts[TSSZ];                               /* timestamp holder */
static const char hextab[]  = "0123456789abcdef";   /* hex digit table */
static const char zSHA256[] = "0000000000000000"    /* zeroed out SHA256 */
                              "0000000000000000"    /* why 4 lines?... */
                              "0000000000000000"
                              "0000000000000000";   /* ...it looks better! */

enum Error {
	ER_NOT_REGULAR,         /* not a regular file error */
	ER_OPENING,             /* error opening file or directory */
	ER_XATTR_OPERATION,     /* error when performing xattr operations */
	ER_GENERIC,             /* generic error */
	ER_FATAL                /* fatal error, should result in program ending */
};

enum FileState {
	FS_OK,                  /* checksum and mtime both match */
	FS_DISRUPTED,           /* file was changed during hash computation */
	FS_CORRUPT,             /* checksum differs while mtime matches */
	FS_MALFORMED,           /* xattrs cannot be read in its intended format */
	FS_TOUCHED,             /* checksum matches but mtime doesn't */
	FS_NEW,                 /* file has no prior checksum or mtime stored */
	FS_OUTDATED,            /* checksum differs and mtime is newer */
	FS_BACKDATED,           /* checksum differs and mtime is older */
	FS_REMOVED,             /* xattrs have been removed from file */
	FS_REMOVED_ERROR,       /* removal of xattrs failed */
	FS_ERROR                /* general error */
};

typedef struct DuplicateNode {
	unsigned char           cs[SHA256_BYTES];   /* checksum */
	char                    fn[PATH_MAX + 1];   /* file name */
	struct DuplicateNode   *next;               /* reference pointer to next dup */
} dn_t;

/* extended attribute/metadata structure */
typedef struct ExtendedAttribute {
	unsigned long long      s;                  /* seconds */
	unsigned long           ns;                 /* nanoseconds */
	char                    hex[CSSZ + 1];      /* checksum (hex) */
	unsigned char           cs[SHA256_BYTES];   /* checksum (binary) */
	int                     tcmp;               /* timespan comparison holder */
} xa_t;

/* arguments collected from command line */
struct Arguments {
	bool                    dryrun;             /* make a dry run */
	int                     quiet;              /* level of quietness */
	bool                    recursive;          /* indicate to open and walk directories */
	bool                    remove;             /* remove xattrs */
	bool                    summarize;          /* show summery at end of program */
	bool                    duplicates;         /* use xattrs to find duplicates */
} arg;

/* counters */
struct Counters {
	int                     errs;               /* all errors */
	int                     errNotRegular;      /* not a regular file errors  */
	int                     errOpening;         /* errors opening file or directory */
	int                     errWritingXattr;    /* errors when performing xattr operations */
	int                     errGeneric;         /* generic errors */

	int                     total;              /* all non error counters */
	int                     ok;                 /* checksum and mtime both match */
	int                     disrupted;          /* file was changed during hash computation */
	int                     corrupt;            /* checksum differs while mtime matches */
	int                     malformed;          /* xattrs cannot be read in its intended format */
	int                     outdated;           /* checksum differs and mtime is newer */
	int                     backdated;          /* checksum differs and mtime is older */
	int                     touched;            /* checksum matches but mtime doesn't */
	int                     new;                /* file has no prior checksum or mtime stored */
} cnt;

/* global variables */
struct Global {
	const char             *prg;                /* program name */
	dn_t                   *dup_head;           /* head of Duplicate nodes */
} gl;

/**
 * Converts binary data into hex format.
 *
 * @param bin               The binary data
 * @param len               The size of the data
 * @param hex               The result of the converted data
 */
static void bin2hex(unsigned char *bin, size_t len, char *hex);

/**
 * Compares two checksums (binary SHA256) to see if they are equal or not.
 *
 * @param cs1               The first checksum
 * @param cs2               The second checksum
 *
 * @returns                 Value < 0 to indicate that cs1 is less than cs2
 * @returns                 Value > 0 to indicate that cs1 is greater than cs2
 * @returns                 Value = 0 to indicate that cs1 is equal to cs2
 */
static int cscmp(const unsigned char *cs1, const unsigned char *cs2);

/**
 * Recursively deallocates the memory previously allocated by a linked
 * duplicate list.
 *
 * @param head_ref          Reference to first (initial call) or current node
 */
static void dup_free(dn_t **head_ref);

/**
 * Divides the nodes in the given linked list into front and back halves
 * referenced by the two reference pointers. If the length is odd, the extra
 * node ends by design up in the front list. As list length is unknown a double
 * step pointer is used together with a single step pointer resulting in
 * the single step pointer referencing the midpoint of the list as the double
 * step pointer reaches the end of the list.
 *
 * @param src               Pointer to the source list to be split
 * @param front_ref         Reference pointer to the resulting front list
 * @param back_ref          Retrieves pointer to the resulting back list
 */
static void dup_front_back_split(dn_t *src, dn_t **front_ref, dn_t **back_ref);

/**
 * Recursively performs the merge sort by first splitting lists, starting from
 * the head of the list, into length of 1 or 0, then by comparing nodes in
 * lists merges them together into one list once again.
 *
 * see: https://en.wikipedia.org/wiki/Merge_sort
 *
 * @param head_ref          Reference to head of each list
 */
static void dup_merge_sort(dn_t **head_ref);

/**
 * Print a linked list of duplicate nodes by first sorting it to then iterate
 * and comparing the current node with the previous to identify duplicates of
 * file content.
 *
 * @param head_ref          Reference to head of list to print
 *
 * @returns                 Returns EXIT_SUCCESS when done
 */
static int dup_print(dn_t **head_ref);

/**
 * Inserts a new node at the beginning of the linked duplicate list with
 * attached filename and checksum from extended file attribute.
 *
 * @param head_ref          Reference to head of linked list
 * @param fn                Name of the file
 * @param fd                File descriptor of the file
 * @param xa                Extended attribute of file
 */
static void dup_push(dn_t **head_ref, const char *fn, int fd, xa_t *xa);

/**
 * Recursively performs the sorting part of the merge sort, where checksums of
 * nodes are compared from two lists being merged together based on comparison.
 *
 * see: https://en.wikipedia.org/wiki/Merge_sort
 *
 * @param a                 One of the lists being compared
 * @param b                 The other list being compared to the first one
 *
 * @returns                 The resulting list from the sorted merge
 */
static dn_t* dup_sorted_merge(dn_t *a, dn_t *b);

/**
 * Retrieves and reports error types and messages. If error type is of fatal
 * nature, the program will exit with EXIT_FAILURE.
 *
 * @param er                The type of error (see enum Error above)
 * @param fmt               The output format of the error message
 * @param ...               Various arguments to output in the message
 */
static void error(enum Error er, const char *fmt, ...);

/**
 * Final stage of processing a file which returns the state of the file being
 * processed meanwhile metadata holders are being populated.
 *
 * @param fd                File descriptor of the file
 * @param xa_s              Metadata holder of (stored) extended attributes
 * @param xa_a              Metadata holder of (actual) extended attributes
 *
 * @returns                 The file state of the file being processed
 */
static enum FileState file_state(int fd, xa_t *xa_s, xa_t *xa_a);

/**
 * Processes arguments (file names) either from command line or from directory
 * output.
 *
 * @param fn                Name of the file
 */
static void process_argument(const char *fn);

/**
 * Processes directories if stated to do so (see arg.recursive) by looping
 * through their content (children).
 *
 * @param fn                Name of the directory
 */
static void process_directory(const char *fn);

/**
 * Processes files and returns either error messages or other output if stated
 * to do so (see arg.quiet) based on their file state (see enum FileState
 * above).
 *
 * @param fn                Name of the file
 */
static void process_file(const char *fn);

/**
 * Reports on files and their file state.
 *
 * @param fn                Name of the file
 * @param st                The state of the file
 */
static void report(const char *fn, enum FileState st);

/**
 * Compute and retrieve metadata for the checksum based on file content
 * together with the date modified.
 *
 * @param fd                File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred
 */
static bool xa_compute(int fd, xa_t *xa);

/**
 * Computes the hash checksum of a file based on the SHA256 hashing algorithm.
 *
 * @param fd                File descriptor of the file
 * @param cs                Pointer for the checksum
 */
static void xa_hash(int fd, unsigned char *cs);

/**
 * Reads and retrieves extended attributes as metadata from a file.
 *
 * @param fd                File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred (malformed data)
 */
static bool xa_read(int fd, xa_t *xa);

/**
 *  Writes, or removes if stated (see arg.remove), extended attributes as
 *  metadata to a file.
 *
 * @param fd               File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred (write error)
 */
static bool xa_write(int fd, const xa_t *xa);

#endif /* !ACST_H */
